/*
    fractal flame generation package
    Copyright (C) 1992, 1993  Scott Draves <spot@cs.cmu.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rect.h"

#ifdef mach
#  include <mach.h>
#endif


/* for batch
 *   interpolate
 *   compute colormap
 *   for subbatch
 *     compute samples
 *     buckets += cmap[samples]
 *   accum += time_filter[batch] * log(buckets)
 * image = filter(accum)
 */


typedef short bucket[3];
typedef long  abucket[3];

#define MAXBUCKET (1<<30)



/* allow this many iterations for settling into attractor */
#define FUSE 15

/* clamp spatial filter to zero at this std dev (2.5 ~= 0.0125) */
#define FILTER_CUTOFF 2.5

/* should be MAXBUCKET / (OVERSAMPLE^2) */
#define PREFILTER_WHITE (MAXBUCKET>>4)

#define SUB_BATCH_SIZE 10000

#define bump_no_overflow(dest, delta, type) { \
   type tt_ = dest + delta;                    \
   if (tt_ > dest) dest = tt_;                  \
}

/* sum of entries of vector to 1 */
void normalize_vector(v, n)
   real *v;
   int n;
{
   real t = 0.0;
   int i;
   for (i = 0; i < n; i++)
      t += v[i];
   t = 1.0 / t;
   for (i = 0; i < n; i++)
      v[i] *= t;
}


void render_rectangle(spec, out, out_width, field)
   frame_spec *spec;
   unsigned char *out;
   int out_width;
{
   int i, j, k, nsamples, nbuckets, batch_size, batch_num, sub_batch;
   bucket  *buckets;
   abucket *accumulate;
   point *points;
   real *filter, *temporal_filter, *temporal_deltas;
   real bounds[4], size[2], ppux, ppuy;
   int image_width, image_height;    /* size of the image to produce */
   int width, height;               /* size of histogram */
   int filter_width;
   int oversample = spec->cps[0].spatial_oversample;
   int nbatches = spec->cps[0].nbatches;
   bucket cmap[CMAP_SIZE];

   image_width = spec->cps[0].width;
   if (field) {
      image_height = spec->cps[0].height / 2;
      if (field == field_odd)
	 out += 3 * out_width;
      out_width *= 2;
   } else
      image_height = spec->cps[0].height;

   if (1) {
      real t;
      filter_width = (2.0 * FILTER_CUTOFF * oversample *
		      spec->cps[0].spatial_filter_radius);
      /* make sure it has same parity as oversample */
      if ((filter_width ^ oversample) & 1)
	 filter_width++;

      filter = (real *) malloc(sizeof(real) * filter_width * filter_width);
      /* fill in the coefs */
      for (i = 0; i < filter_width; i++)
	 for (j = 0; j < filter_width; j++) {
	    real ii = ((2.0 * i + 1.0) / filter_width - 1.0) * FILTER_CUTOFF;
	    real jj = ((2.0 * j + 1.0) / filter_width - 1.0) * FILTER_CUTOFF;
	    if (field)
	       jj *= 2.0;
	    filter[i + j * filter_width] =
	       F(exp)(-2.0 * (ii * ii + jj * jj));
	 }

      normalize_vector(filter, filter_width * filter_width);
   }
   temporal_filter = (real *) malloc(sizeof(real) * nbatches);
   temporal_deltas = (real *) malloc(sizeof(real) * nbatches);
   if (nbatches > 1) {
      real t;
      /* fill in the coefs */
      for (i = 0; i < nbatches; i++) {
	 t = temporal_deltas[i] = (2.0 * ((real) i / (nbatches - 1)) - 1.0)
	    * spec->temporal_filter_radius;
	 temporal_filter[i] = F(exp)(-2.0 * t * t);
      }

      normalize_vector(temporal_filter, nbatches);
   } else {
      temporal_filter[0] = 1.0;
      temporal_deltas[0] = 0.0;
   }

#if 0
   for (j = 0; j < nbatches; j++)
      fprintf(stderr, "%4f %4f\n", temporal_deltas[j], temporal_filter[j]);
   fprintf(stderr, "\n");
#endif
   
   /* the number of additional rows of buckets we put at the edge so
      that the filter doesn't go off the edge */
   if (1) {
      real t0, t1, shift;
      int gutter_width;
      ppux = spec->cps[0].pixels_per_unit;
      ppuy = field ? (ppux / 2.0) : ppux;
      switch (field) {
       case field_both: shift =  0.0; break;
       case field_even: shift = -0.5; break;
       case field_odd:  shift =  0.5; break;
      }
      shift = shift / ppux;
      gutter_width = (filter_width - oversample) / 2;
      height = oversample * image_height + 2 * gutter_width;
      width  = oversample * image_width  + 2 * gutter_width;
      t0 = (real) gutter_width / (oversample * ppux);
      t1 = (real) gutter_width / (oversample * ppuy);
      bounds[0] = spec->cps[0].corner[0] - t0;
      bounds[1] = spec->cps[0].corner[1] - t1 + shift;
      bounds[2] = spec->cps[0].corner[0] + image_width  / ppux + t0;
      bounds[3] = spec->cps[0].corner[1] + image_height / ppuy + t1 + shift;
      size[0] = 1.0 / (bounds[2] - bounds[0]);
      size[1] = 1.0 / (bounds[3] - bounds[1]);
   }

   nbuckets = width * height;
   nsamples = (int) (spec->cps[0].sample_density * nbuckets /
		     (oversample * oversample));
/*    fprintf(stderr, "nsamples = %d\n", nsamples); */
/*    fprintf(stderr, "nbuckets = %d\n", nbuckets); */
   
   batch_size = nsamples / spec->cps[0].nbatches;
   if (1) {
     static char *last_block = NULL;
     static int last_block_size = 0;
     int memory_rqd = (sizeof(bucket) * nbuckets +
		       sizeof(abucket) * nbuckets +
		       sizeof(point) * SUB_BATCH_SIZE);
     int Mb = memory_rqd / (1024 * 1024);
     if (memory_rqd > last_block_size) {
#ifdef mach
       vm_address_t address;
       kern_return_t err;
       if (last_block != NULL)
	 fprintf(stderr, "should vm_free but don't know how\n");
       err = vm_allocate(task_self(), &address, (vm_size_t) memory_rqd,
			 (boolean_t) 1);
       if (KERN_SUCCESS != err) {
	 fprintf(stderr, "render_rectangle: cannot vm_allocate %d bytes.\n", memory_rqd);
	 mach_error("vm_allocate", err);
	 exit(1);
       } else
	 fprintf(stderr, "render_rectangle: vm_allocated %dMb.\n", Mb);
	 
       last_block = (char *) address;
#else
       if (last_block != NULL)
	  free(last_block);
       last_block = (char *) malloc(memory_rqd);
       if (NULL == last_block) {
	  fprintf(stderr, "render_rectangle: cannot malloc %d bytes.\n", memory_rqd);
	  exit(1);
       } else
	  fprintf(stderr, "render_rectangle: mallocked %dMb.\n", Mb);

#endif
       last_block_size = memory_rqd;
     }
     buckets = (bucket *) last_block;
     accumulate = (abucket *) (last_block + sizeof(bucket) * nbuckets);
     points = (point *)  (last_block + (sizeof(bucket) + sizeof(abucket)) * nbuckets);
   }

   bzero((char *) accumulate, sizeof(abucket) * nbuckets);
   for (batch_num = 0; batch_num < nbatches; batch_num++) {
      real batch_time;
      control_point cp;
      bzero((char *) buckets, sizeof(bucket) * nbuckets);
      batch_time = spec->time + temporal_deltas[batch_num];

      /* interpolate and get a control point */
      interpolate(spec->cps, spec->ncps, batch_time, &cp);

      /* compute the colormap entries.  the input colormap is 256 long with
	 entries from 0 to 1.0 */
      for (j = 0; j < CMAP_SIZE; j++)
	 for (k = 0; k < 3; k++) {
#if 1
	    cmap[j][k] = (int) (cp.cmap[(j * 256) / CMAP_SIZE][k] *
				cp.white_level);
#else
	    /* monochrome if you don't have any cmaps */
	    cmap[j][k] = cp.white_level;
#endif
	 }

      for (sub_batch = 0;
	   sub_batch < batch_size;
	   sub_batch += SUB_BATCH_SIZE) {

	 /* generate a sub_batch_size worth of samples */
	 points[0][0] = random_uniform11();
	 points[0][1] = random_uniform11();
	 points[0][2] = random_uniform01();
	 iterate(&cp, SUB_BATCH_SIZE, FUSE, points);
	 
	 
	 /* merge them into buckets, looking up colors */
	 for (j = 0; j < SUB_BATCH_SIZE; j++) {
	    int k, color_index;
	    real *p = points[j];
	    bucket *b;
	    if (p[0] < bounds[0] ||
		p[1] < bounds[1] ||
		p[0] > bounds[2] ||
		p[1] > bounds[3])
	       continue;
	    color_index = (int) (p[2] * CMAP_SIZE);
	    if (color_index < 0) color_index = 0;
	    else if (color_index > (CMAP_SIZE-1))
	       color_index = CMAP_SIZE-1;
	    b = buckets +
	       (int) (width * (p[0] - bounds[0]) * size[0]) +
		  width * (int) (height * (p[1] - bounds[1]) * size[1]);
	    for (k = 0; k < 3; k++)
	       bump_no_overflow(b[0][k], cmap[color_index][k], short);
	 }
      }
      
      if (1) {
	 real k1 = (cp.contrast * cp.brightness *
		    PREFILTER_WHITE * 268.0 *
		    temporal_filter[batch_num]) / 256;
	 real area = image_width * image_height / (ppux * ppuy);
	 real k2 = (oversample * oversample * nbatches) /
	    (spec->cps[0].contrast * area * spec->cps[0].white_level *
	     spec->cps[0].sample_density);
 

	 /* log intensity in hsv space */
	 for (j = 0; j < height; j++)
	    for (i = 0; i < width; i++) {
	       abucket *a = accumulate + i + j * width;
	       bucket *b = buckets + i + j * width;
	       int t;
	       real c[3];
	       c[0] = (real) b[0][0];
	       c[1] = (real) b[0][1];
	       c[2] = (real) b[0][2];
	       if (0.0 == c[0] &&
		   0.0 == c[1] &&
		   0.0 == c[2])
		  continue;
	       rgb2hsv(c, c);
	       c[2] = k1 * F(log)(1.0 + c[2] * k2);
	       hsv2rgb(c, c);

	       bump_no_overflow(a[0][0], c[0] + 0.5, int);
	       bump_no_overflow(a[0][1], c[1] + 0.5, int);
	       bump_no_overflow(a[0][2], c[2] + 0.5, int);
	    }
      }
   }
   /*
    * filter the accumulation buffer down into the image
    */
   if (1) {
      int x, y;
      real t[3], k;
      real g = 1.0 / spec->cps[0].gamma;
      y = 0;
      for (j = 0; j < image_height; j++) {
	 x = 0;
	 for (i = 0; i < image_width; i++) {
	    int ii, jj, a;
	    unsigned char *p;
	    t[0] = t[1] = t[2] = 0.0;
	    for (ii = 0; ii < filter_width; ii++)
	       for (jj = 0; jj < filter_width; jj++) {
		  real k = filter[ii + jj * filter_width];
		  abucket *a = accumulate + x + ii + (y + jj) * width;
		  
		  t[0] += k * a[0][0];
		  t[1] += k * a[0][1];
		  t[2] += k * a[0][2];
	       }
	    p = out + 3 * (i + j * out_width);
	    a = 256.0 * F(pow)((real) t[0] / PREFILTER_WHITE, g) + 0.5;
	    if (a < 0) a = 0; else if (a > 255) a = 255;
	    p[0] = a;
	    a = 256.0 * F(pow)((real) t[1] / PREFILTER_WHITE, g) + 0.5;
	    if (a < 0) a = 0; else if (a > 255) a = 255;
	    p[1] = a;
	    a = 256.0 * F(pow)((real) t[2] / PREFILTER_WHITE, g) + 0.5;
	    if (a < 0) a = 0; else if (a > 255) a = 255;
	    p[2] = a;
	    x += oversample;
	 }
	 y += oversample;
      }
   }
   
   free(filter);
}
